/* ************************************************** *
 *						      *
 *  Code to record / play back X events in xscript    *
 *                                                    *
 *	(c) Copyright 1993, Oliver Jones	      *
 *						      *
 * ************************************************** */

#include "scope.h"
#include "record.h"
#include "x11.h"
#include <X11/Xlib.h>



/* ************************************************************ */
/*								*/
/*								*/
/* ************************************************************ */

/* stored script line */

static char script_line [256];


/* ************************************************************ */
/*								*/
/*								*/
/* ************************************************************ */

static FILE *
OpenFile (name, priv)
     char * name;
     char * priv;
{
  register FILE * file;
  register int fid;

  /* open file */
  file = fopen (name, priv);

  /* deal with failure.  snarky messages are OK, we've pretested files */
  if (file == 0) {
    perror (file);
    panic ("OpenFile failure.\n");
  }

  /*
   * set up file descriptor in this program's table.
   * Notice that UsingFD knows not to hammer a read-handler
   * if the handler address passed in is 0
   */
  fid = fileno(file);
  UsingFD (fid, 0);
  return file;
}


/* ************************************************************ */
/*								*/
/*								*/
/* ************************************************************ */
/*
 *  TimeDilation expresses the time dilation in percent.
 *  if TimeDilation = 200, then we want the thing to play back
 *  twice as fast, so we want to cut all times in half.
 */

unsigned long int
Dilate (deltat)
     unsigned long int deltat;
{
  unsigned long int result;
  result = deltat;
  result = ((deltat * 100) + ((TimeDilation/2) - 1)) / TimeDilation;
  return result;
}

/* ************************************************************ */
/*								*/
/*								*/
/* ************************************************************ */


static long int
NextScriptEvent (buf, n)
     char * buf;
     long int n;
{
  /* 
   * This function reads the next script event
   * and returns it in a buffer.
   */
  char * result;
  int goodline = 0;
  int event;

  if (PlaybackFile == 0) {
    /* open up the recording file if necessary, first time through */
    PlaybackFile = OpenFile (PlaybackFileName,  /* name to open */
			     "r");            /* IO mode */
    /* mark playback in progress */
    Playback = PLAYBACK_ACTIVE;
  }


  /* skip blanks and comments */

  while (goodline ==0) {
    buf[0] = '\0';
    result = fgets (buf, n, PlaybackFile);
    if (result == 0) return -1;
    
    /* skip comment-type lines */
    if (buf[0] != '\0') {
      event = ascii_to_event_type (buf);
      if (event != 0 ) {
	goodline = 1;
      }
      else {
	fprintf (stderr, "commment %d %s", event, buf);
      }
    }
    
  }
  return strlen(buf);

}


/* ************************************************************ */
/*								*/
/*								*/
/* ************************************************************ */

void (*Resource_Handler) () = 0;     /* handler to call */
void * Resource_userdata    = 0;     /* user data to pass */
XID    Resource_resource    = 0;
TimerID Resource_timeout    = 0;     /* timeout when resource doesn't appear */
int    Resource_index       = 0;     /* resource index within fd  */
FD     Resource_fd          = 0;     /* resource file descriptor */


/*
 *  this timer function is called when 
 *  the wait for a window to become mapped 
 *  runs out...
 *  It issues a warning message, then proceeds as
 *  if the window had, in fact, appeared.
 *
 *  Under some failure conditions, this causes an
 *  error cascade.
 */

/* ARGSUSED */
static void 
ResourceNeverAppeared  (timer, window)
     TimerID timer;
     XID     window;
{
  fprintf (stderr, "Playback failure\n");
  fprintf (stderr, "  Timed out waiting for window %8.8x\n", window );
  fprintf (stderr, "  %s\n", script_line);
  Errorcount ++;

  /* call the wait handler, attempting to carry on with the run */
  if (Resource_Handler) 
    (*Resource_Handler) (Resource_resource, Resource_userdata);

  /* clear off the waiting flag in the resource status map */
  if (Resource_index < MAXRESOURCESTATUS) {
    CS[Resource_fd].ResourceStatus[Resource_index] &= ~RES_Waiting;
  }

  /* Restore the resource-wait information to initial state */
  Resource_Handler          = 0; 
  Resource_userdata         = 0;
  Resource_resource         = 0;
  Resource_timeout          = 0;
  Resource_index            = 0;
  Resource_fd               = 0;
  
  return;
}

/*
 * this function can be called when somebody
 * wants to wait for a window to appear.
 */

void 
WaitForWindowMapped (fd, resource, handler, user_data)
     int fd;
     XID resource;
     void (*handler)();
     void * user_data;
{
  int res_index = resource & CS[fd].ResourceIdMask;
  
  if (Resource_Handler) panic ("already waiting for a resource....only one allowed!");

  if (res_index >= MAXRESOURCESTATUS) {
    /* beyond the resource count; just assume it's available */
    if (handler) 
      (*handler) (resource, user_data);
    return;
  }
  
  if (CS[fd].ResourceStatus[res_index] == RES_Mapped) {
    (*handler) (resource, user_data);
    return;
  }
  
  CS[fd].ResourceStatus[res_index] |= RES_Waiting;
  Resource_Handler  = handler;
  Resource_userdata = user_data;
  Resource_resource = resource;
  Resource_timeout  = CreateTimer (2000, ResourceNeverAppeared, resource);
  Resource_index    = res_index;
  Resource_fd       = fd;

}
  

/*
 * this routine gets called from the event formatters
 * in print11.c
 * it is used to maintain event status.
 */


void 
WindowStatusAnnounce (fd, buf, n, status)
     FD fd;
     unsigned char * buf;
     int n;
     ResStatus status;
{
  XID resource = ILong (&buf[n]);
  int res_index = resource & CS[fd].ResourceIdMask;
  int waiting = 0;
  ResStatus oldstatus;

  if (res_index < MAXRESOURCESTATUS) {
    oldstatus = CS[fd].ResourceStatus[res_index];
    if (oldstatus & RES_Waiting)  waiting=1;
    CS[fd].ResourceStatus[res_index] = status | (oldstatus & RES_Waiting);
  }

  /* if we were waiting for the resource and it is now mapped ... */
  if (waiting & (status == RES_Mapped)) {

    /* Blast the wait timeout */
    if (Resource_timeout) 
      DeleteTimer (Resource_timeout);

    /* call the resource handler, if any */
    if (Resource_Handler) 
      (*Resource_Handler) (Resource_resource, Resource_userdata);
    
    /* zap the waiting bit in the resource map */
    if (Resource_index < MAXRESOURCESTATUS) {
      CS[Resource_fd].ResourceStatus[Resource_index] &= ~RES_Waiting;
    }

    /* Restore the resource-wait information to initial state */
    Resource_Handler          = 0; 
    Resource_userdata         = 0;
    Resource_resource         = 0;
    Resource_timeout          = 0;
    Resource_index            = 0;
    Resource_fd               = 0;
  }
  
  return;
  
}
  
/* ************************************************************ */
/*								*/
/*								*/
/* ************************************************************ */

/*ARGSUSED*/
AnnouncePlaybackError (dpy, erev)
     Display * dpy;
     XErrorEvent * erev;
{
  
  int type = ascii_to_event_type (script_line);
  if (type != EnterNotify) {
    fprintf (stderr, "Playback failure\n");
    fprintf (stderr, "  Unable to move pointer into window %8.8x\n",
 	     erev->resourceid);
    fprintf (stderr, "  %s\n", script_line);
    Errorcount ++;
  }
}


/* ************************************************************ */
/*								*/
/*								*/
/* ************************************************************ */

/*ARGSUSED*/
SendPlaybackEvent (timer, fd)
     TimerID timer;
     int fd;
{
  unsigned char buf [32];  /* event buffer */
  struct timeval tv;
  FD pairfd = FDPair(fd);

  unsigned short int seq;
  Time timestamp, t;

  /* obtain the sequence number from the client-side CS structure */
  if (pairfd >= StaticMaxFD) pairfd = -1;
  if (pairfd >= 0)  seq = CS[pairfd].SequenceNumber;
  
  /* compute timestamp based on previously noted time offset */
  gettimeofday (&tv, 0);
  t = tv.tv_sec * 1000;
  t += tv.tv_usec / 1000;
  timestamp = t + CS[fd].TimeOffset;
  
  /* it's time to send the event to the program under test..  */
  /* convert it to a wire-format event buffer */
  if (ascii_to_event_buf (script_line,  buf, seq, timestamp)) {
    WriteToClient (fd, buf, 32);
  }

  /* read the next event from the script file */
  ReadPlaybackEvent (timer, fd);

}

/* ************************************************************ */
/*								*/
/*								*/
/* ************************************************************ */

/*ARGSUSED*/
WarpPlaybackEvent (session_window, fd)
     Window  session_window;
     int fd;
{
  /*
   * This function decodes part of a stored script_line,
   * and warps the server pointer to the selected
   * position.  It is called when the window is known to be mapped.
   * This function ends by registering a callback to 
   * "SendPlaybackEvent" for 20 ms later.
   */

  extern Display * Dpy;
  Window saved_window;
  unsigned int x,y;
  static int first = 0;
  
  if (first == 0) {
    (void) XSetErrorHandler (AnnouncePlaybackError);
    first = 1;
  }
  
  ascii_to_x_y_window (script_line, &x, &y, &saved_window);

  XWarpPointer (Dpy, None, session_window, 0,0,0,0,x,y);

  XSync (Dpy, False);

  CreateTimer (20L, SendPlaybackEvent, fd);
   
}

/* ************************************************************ */
/*								*/
/*								*/
/* ************************************************************ */

/*ARGSUSED*/
TranslatePlaybackEvent (timer, fd)
     TimerID timer;
     int fd;
{
  /*
   * This function decodes part of a stored script_line,
   * and waits for the named window to appear.
   */

  Window saved_window;
  Window session_window;
  unsigned int x,y;

  ascii_to_x_y_window (script_line, &x, &y, &saved_window);
  session_window = WindowSavedToSession (saved_window);
  WaitForWindowMapped (fd, session_window, WarpPlaybackEvent, fd);
}


/* ************************************************************ */
/*								*/
/*								*/
/* ************************************************************ */

/*ARGSUSED*/
ReadPlaybackEvent (timer, fd)
     TimerID timer;
     int fd;

{
  /*
   * This function reads an event from the script file.
   * It then establishes a timer call back to 
   * activate WarpPlaybackEvent at the appropriate time
   * (based on the elapsed time specified in the script file.)
   *
   * If NextScriptEvent comes back = -1, it's the end of
   *   the script file;  forget about further timer callbacks.
   * However, don't stop the xscript pseudo-server, because
   *   it should stay running until the client-under-test
   *    disconnects.
   */

  long int linelen = NextScriptEvent (script_line, 256);
  unsigned long int delay, delay2;

  if (linelen < 0) {
    /* end of file ... mark playback as complete */
    Playback = PLAYBACK_COMPLETE;
  }
  else {
    /* process next event from file. */
    delay = ascii_to_delay (script_line);
    switch (ascii_to_event_type (script_line)) {
    case ButtonPress:
      /*
       *  to avoid fouling up double clicks,
       *  don't mess around with delay times before ButtonPress < 250 ms 
       */
      if (delay < DoubleClickTime)  {
	/* delay is shorter than doubleclick time, don't touch it */
	delay2 = delay;
      }
      else {
	/*
         * original delay is longer than doubleclick time, 
         * don't let recomputed delay end up 
         * shorter than doubleclick time
         */
	delay2 = Dilate (delay);
	if (delay2 < DoubleClickTime) delay2 = delay;
      }
      break;

    default:
      /* it's OK to compress other event delays more-or-less arbitrarily */
      delay2 = Dilate (delay);
      break;
    }

    /* subtract delay time we'll use for waiting for motion */
    if (delay2 > 20) delay2 -= 20;
    else delay2 = 0;

    if (delay2 < 5) delay2 = 5;
    debug(128,(stderr,"%d=delay %s\n", delay2, script_line));
    CreateTimer (delay2, TranslatePlaybackEvent, fd);
  }
   
}

/* ************************************************************ */
/*								*/
/*								*/
/* ************************************************************ */

/*ARGSUSED2*/
RecordEvent (fd, buf, n)
     FD fd;
     unsigned char *buf;
     long    n;
{
  
  register short type = IByte (&buf[0]);  /* event type */
  
  if (RecordFile == 0) {
    /* open up the recording file if necessary, first time through */
    RecordFile = OpenFile (RecordFileName,  /* name to open */
			   "w");            /* IO mode */
  }
  
  /* high-order bit means SendEvent generated */
  if (type & 0x80) {
    type = type & 0x7F;
  }

  /*
   * note that we can't cast stuff to
   * the XEvent structure here, because
   * the information is in a message buffer,
   * not correctly unpacked and byteswapped.
   */
  switch(type)        
  {
  case ButtonPress : 
  case ButtonRelease :
  case KeyPress : 
  case KeyRelease :
  case MotionNotify :
  case EnterNotify :
    /* process the event */
    event_to_ascii (buf, RecordFile);
    break;
  default:
    /*EMPTY*/ /* ignore the event */
    break;
  }
}


