.TH XSCRIPT 1 "28 FEB 1993" "X Version 11"
.SH NAME
xscript - X Window System Script Recording and Playback
.SH SYNOPSIS
.B xscript
[ options ] ...
.SH DESCRIPTION
.I xscript
is a program which sits between an X client and an X server, 
monitors the events flowing from the server to the client,
and records events in a script file. The script file name
for recording is specified using the -r option.
.I xscript 
is useful as part of a regression test
system, as it can automate the playing of
fixed sequences of operations into an 
application program based on X.
.I xscript
can play back events from a script 
file (specified using the -p option) to a client.   
.PP
.I xscript 
can detect application-program failures during playback.
To do so, it detects the failure of windows to 
appear as necessary for playback, and
reports errors when windows do not appear within one second
of when they are expected.  A failure of a window to appear is
symptomatic of several classes of application failures, including
failure of a popup to appear and an application crash.
.PP
Upon completing a playback session which generated errors, xscript
exits with nonzero status.
.PP
.I xscript
is based on James L. Peterson's \fIxscope\fP program.
.PP
\fIxscript\fP functions as a protocol interposition process.  It 
attaches to the X server as if it were a client, and a client attaches
to \fIxscript\fP as if it were an X server.  This allows \fIxscript\fP 
to be used for testing standard unmodified client application programs.
.PP
By adjusting the host and/or display number, the client can be made to
attach to \fIxscript\fP instead of the X server.
.PP
	server <-----> xscript <-----> client
.PP
All bytes from the server are sent to \fIxscript\fP which sends them on to
the client.
All bytes from the client are sent to \fIxscript\fP which passes them on
to the server.
\fIxscript\fP is transparent to both the server and the client.
.SH OPTIONS
.PP
.TP 10
.B \-r<script-file-name>
Requests the recording of a session and 
defines the script file to be written during
recording.
.PP
.TP 10
.B \-p<script-file-name>
Requests the playback of a session and 
defines the script file to be played back.  
The beginning of playback is deferred until the client program
attaches.
.PP
.TP 10
.B \-t<time-contraction-percent>
Specifies the percent time contraction to be used
during playback.  -d50 plays back a session at 50% speed,
and -d200 plays it back at double speed.  Changing playback
speeds may foul up the client's detection of double clicks.
The default time contraction is 100 percent (1:1 playback:record).
.PP
.TP 10
.B \-c<double-click-time>
Specifies the double-click-time in milliseconds.  During playback when
time is contracted or expanded, ButtonPress events with short delays
from the previous event are handled specially so that they do not get
misrecognized by the client.
.PP
.TP 10
.B \-w<wait-before-playback>
Specifies the number of seconds to wait between the time
the client attaches and the first event is played back.  
This allows complex clients to complete drawing before
playback begins.  The default is five seconds.
.PP
.TP 10
.B \-d<display>
Defines the display number.
The display number is added to the input and output port to give the
actual ports which are used by \fIxscript\fP.
.PP
.TP 10
.B \-h<host>
Specifies the hostname on which \fIxscript\fP should look for its X server.
.PP
.PP
.TP 10
.B \-h
Prints a help message.
.PP
.TP 10
.B \-o<output-port>
Determines the port that \fIxscript\fP will use to connect to the server
(default is 0).   For example, if \fIxscript\fP were run with the
options "-hlaser -o0" it would connect to the DISPLAY named "laser:0".
In the normal case it is not necessary to specify \-h or \-o options:
\fIxscript\fP connects, via TCP, to the local host's zeroth display.
.PP
.TP 10
.B \-i<input-port>
Specify the port that \fIxscript\fP will use to take requests from clients
(default is 1).  For example, if \fIxscript\fP were
running on the host named "hobie," 
and the option \-i3 were specified, 
clients for recording or playback would attach to the
DISPLAY named "hobie:3".  Normally, it is not necessary to
specify the \-i option, and clients should be made to connect
via TCP to "localhost:1".
.PP
.TP 10
.B \-q
Quiet output mode.  Gives only the names of requests, replies, errors, and 
events, but does not indicate contents.
.PP
.TP 10
.B \-v<print-level>
Determines the level of verbosity which \fIxscript\fP will provide.
The print-level can be 0 (same as quiet mode), 1, 2, 3, 4.
The larger numbers give more and more output.
For example, a successful setup returns a string which is the name of the
vendor of the X server.
At level 1, the explicit field giving the length of the string is
suppressed since it can be inferred from the string.
At level 2 and above the length is explicitly printed.
.PP
.TP 10
.B \-G<timeout>
Determines the grab timeout.  If any active grab (server,
pointer, or keyboard) originating from the program under
test lasts longer than this time (default 30 seconds) the
xscript program terminates, releasing the grab.  This feature
is handy for use in debugging situations in which grabs deadlock the
server.
.SH OPERATION
When running with \fIxscript\fP, three processes are involved, potentially
all on different machines:
.PP
.TP 5
\fIX server\fP
The X server will run on machine (or host) "A", display "B".
("A" is a machine name; "B" is a display number).
.PP
.TP 5
\fIxscript\fP
\fIxscript\fP must be told where the X server is (what machine and
what display), and, in addition, the \fIport-number\fP on which to listen
for X clients.
The options for \fIxscript\fP are "-h<Xserver-host>" and "-d<display-number>".
In this example, -hA and -dB.
Typically the \fIdisplay-number\fP is not given.
\fIxscript\fP will not try to  connect to the server until the client
connects to \fIxscript\fP.
.PP
.TP 5
\fIX client\fP
The client should connect to \fIxscript\fP rather than to the server.
To avoid changing the code for the client, \fIxscript\fP listens on
the same port as the server for connecting clients.
If the server and \fIxscript\fP are on different machines, this works well.
However, if the server and \fIxscript\fP are on the same machine, this 
creates a port conflict.
To resolve this conflict, \fIxscript\fP can be given a different
input or output port number, as necessary, to avoid the port that
the server is listening to.
The client must connect to this offset port number.
The input port for \fIxscript\fP is set by \fI-i<port-number>\fP;
the output port is set by \fI-o<port-number>.\fP
The default input port is 1; the default output port is 0.
These ports are offset by the X11 base (6000) and the display number.
The client attaches to \fIxscript\fP by changing its display number by the
port offset.
.SH EXAMPLES
.LP
.B "xscript -hbagel -r/tmp/record"
.PP
The X server would be expected to be found 
on "bagel", display 0 (the default).
\fIxscript\fP and the client would both be running on "cleo".
The client program would think it was connecting to an X server on 
"cleo:1".  This X server is actually \fIxscript\fP, which would
record events and pass through requests and events to the real
X server running on "bagel:0".
.PP
  server (bagel:0) <---> xscript -hbagel <---> client -d cleo:1
.PP
As it runs, \fIxscript\fP records input events for the client
to the file /tmp/record.
.LP
.B "xscript -p/tmp/record"
.PP
The X server would be expected to be found 
on the local host, display 0 (the default).
\fIxscript\fP and the client would be running on the same local host.
The client would connect to "localhost:1", the \fIxscript\fP 
program, which in turn would connect to the real X server on
"localhost:0".
.PP
  server (localhost:0) <---> xscript  <---> client -display localhost:1
.PP
As it runs, \fIxscript\fP plays back pre-recorded events for the client
from the file /tmp/record.  If all is well, the recorded session
should be reproduced.
.SH LIMITATIONS
The X Version 11 is the only protocol recognized.
.PP
The error-detection scheme on playback is extremely limited.
.PP
The assumptions made about the similarity between the recorded and played
back client sessions are easy to break.
.PP
Only one client can be recorded or played back in a session.
.PP
The builtin atoms have been wired in directly.
A better solution is to pick them up from a header file.
.PP
There is no code yet to interpret typed commands from the keyboard.
It would be possible for a command language at the keyboard to create
artificial characters to be sent to the server or the client as if they were 
generated  by the opposite side, or to dynamically alter requests or replies.
.SH SEE ALSO
X(1), X11 and PEX Protocol documents
.SH AUTHOR
.PP
James L. Peterson (MCC)
.PP
Oliver Jones
.PP
Copyright 1988, MCC
.PP
Copyright 1993, Shawsheen Software
